@@ -1,5 +1,5 @@
 use hir::known;
-use ide_db::helpers::FamousDefs;
+use ide_db::famous_defs::FamousDefs;
 use stdx::format_to;
 use syntax::{
     ast::{self, edit_in_place::Indent, make, HasArgList, HasLoopBody},
@@ -154,10 +154,11 @@ fn is_ref_and_impls_iter_method(
     let has_wanted_method = ty
         .iterate_method_candidates(
             sema.db,
-            krate,
+            &scope,
             &traits_in_scope,
+            None,
             Some(&wanted_method),
-            |_, func| {
+            |func| {
                 if func.ret_type(sema.db).impls_trait(sema.db, iter_trait, &[]) {
                     return Some(());
                 }
